r"""
ECD
##########################################

Reference:
    Yuqiang Zhou et al. "Modeling Context-aware Features for Cognitive Diagnosis in Student Learning" in KDD 2021.


"""
from ..gd_basemodel import GDBaseModel
import torch.nn as nn
from ..utils.components import PosMLP
import torch
import torch.nn.functional as F

class ECD(GDBaseModel):
    r"""
    ECD

    default_cfg:
        'dnn_units': [512, 256]  # dimension list of hidden layer in prediction layer
        'dropout_rate': 0.5      # dropout rate
        'activation': 'sigmoid'  # activation function in prediction layer
        'disc_scale': 10         # discrimination scale
        'emb_dim': 10           # dimension of q,k
        'con_dim': 1            # dimension of v
    """
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'disc_scale': 10,
        'emb_dim': 10,
        'con_dim': 1,
    }
    def __init__(self, cfg):
        super().__init__(cfg)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.qk_dim = self.modeltpl_cfg["emb_dim"]
        self.con_dim = self.modeltpl_cfg["con_dim"]
        self.batch_size = self.traintpl_cfg['batch_size']

    def build_model(self):
        self.con_ans_num = self.datatpl_cfg['dt_info']['qqq_count']
        self.stu_q = nn.Embedding(self.n_user, self.qk_dim, padding_idx=0)
        self.qqq_group = [torch.tensor(k).to(self.device) for k in self.qqq_group_list]
        self.group_len = len(self.qqq_group)
        self.con_ans_v = nn.Embedding(self.con_ans_num, self.con_dim, padding_idx=0)
        self.con_ans_k = nn.Embedding(self.con_ans_num, self.qk_dim, padding_idx=0)

        self.d_t = nn.Embedding(self.n_user, 1, padding_idx=0)

    def add_extra_data(self, **kwargs):
        self.qqq_group_list = kwargs['qqq_group_list']
        self.Q_mat = kwargs['Q_mat'].to(self.device)
        self.QQQ_list = kwargs['qqq_list'].to(self.device)

    def atten(self, query, key, value):
        v_dim = value.shape[2]
        btch = query.shape[0]
        temp0 = torch.matmul(query, key.transpose(1, 2))
        temp_query_n = query.norm(dim=2).view(btch, -1, 1)
        temp_key_n = key.norm(dim=2).view(btch, 1, -1)
        temp_n = temp_query_n * temp_key_n
        temp_sim = temp0 / temp_n
        sim = F.softmax(temp_sim, dim=2).transpose(1, 2)
        res = torch.matmul(value.transpose(1, 2), sim).transpose(1, 2).view(btch, -1, v_dim)
        return res

    def self_atten(self, query, key, value):
        v_dim = value.shape[2]
        btch = query.shape[0]
        temp0 = torch.matmul(query, key.transpose(1, 2))
        temp_query_n = query.norm(dim=2).view(btch, -1, 1)
        temp_key_n = key.norm(dim=2).view(btch, 1, -1)
        temp_n = temp_query_n * temp_key_n
        temp_sim = temp0 / temp_n
        sim = F.softmax(temp_sim, dim=2).transpose(1, 2)
        res = torch.matmul(value.transpose(1, 2), sim).transpose(1, 2).view(btch, -1, v_dim)
        return res

    def forward(self, stu_id, qqq_ids):
        # before prednet
        group_v = torch.tensor([]).type(torch.FloatTensor).to(self.device)
        group_k = torch.tensor([]).type(torch.FloatTensor).to(self.device)
        btch = stu_id.shape[0]
        stu_q = self.stu_q(stu_id).view(btch, 1, -1)
        for i in range(self.group_len):
            temp_query = stu_q
            temp_idx = qqq_ids[:, self.qqq_group[i]]
            temp_v = self.con_ans_v(temp_idx)
            temp_k = self.con_ans_k(temp_idx)
            temp_value = self.atten(temp_query, temp_k, temp_v)
            temp_key = self.atten(temp_query, temp_k, temp_k)
            group_v = torch.cat([group_v, temp_value], dim=1)
            group_k = torch.cat([group_k, temp_key], dim=1)
        group_value = self.self_atten(group_k, group_k, group_v)
        group_key = self.self_atten(group_k, group_k, group_k)
        temp_query = stu_q
        context_emb = self.atten(temp_query, group_key, group_value)
        context_emb = torch.sigmoid(context_emb.view(btch, self.con_dim))
        dt_w = torch.sigmoid(self.d_t(stu_id)).view(btch, 1)
        return context_emb, dt_w

class ECD_IRT(ECD):
    default_cfg = {
        "a_range": -1.0,  # disc range
        "diff_range": -1.0,  # diff range
        "fix_a": False,
        "fix_c": True,
        'loss_weight': [4, 0],
    }
    def build_model(self):
        super().build_model()
        self.theta = nn.Embedding(self.n_user, 1)  # student ability
        self.a = 0.0 if self.modeltpl_cfg['fix_a'] else nn.Embedding(self.n_item, 1)  # exer discrimination
        self.b = nn.Embedding(self.n_item, 1)  # exer difficulty

    @staticmethod
    def irf(theta, a, b,  D=1.702):
        return 1 / (1 + torch.exp(-D * a * (theta - b)))

    def forward(self, stu_id, exer_id):
        items_Q_mat = self.Q_mat[exer_id]
        stus_qqq = self.QQQ_list[stu_id]
        theta_inner = self.theta(stu_id)
        a = self.a(exer_id)
        b = self.b(exer_id)
        theta_context, dt_w = super().forward(stu_id, stus_qqq)
        theta_all = theta_context * dt_w + theta_inner * (1 - dt_w)
        if self.modeltpl_cfg['diff_range'] is not None:
            b = self.modeltpl_cfg['diff_range'] * (torch.sigmoid(b) - 0.5)
        if self.modeltpl_cfg['a_range'] is not None:
            a = self.modeltpl_cfg['a_range'] * torch.sigmoid(a)
        else:
            a = F.softplus(a) # 让区分度大于0，保持单调性假设
        if torch.max(theta_inner != theta_inner) or torch.max(a != a) or torch.max(b != b):  # pragma: no cover
            raise ValueError('ValueError:theta,a,b may contains nan!  The diff_range or a_range is too large.')
        pre_all = self.irf(theta_all, a, b)
        pre_context = self.irf(theta_context, a, b)
        pre_inner = self.irf(theta_inner, a, b)
        if self.training:
            return pre_all, pre_context, pre_inner
        else:
            return pre_all

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd_all, pd_context, pd_inner = self(stu_id, exer_id)
        pd_all, pd_context, pd_inner = pd_all.flatten(), pd_context.flatten(), pd_inner.flatten()
        loss_all = F.binary_cross_entropy(input=pd_all, target=label)
        loss_context = F.binary_cross_entropy(input=pd_context, target=label)
        loss_inner = F.binary_cross_entropy(input=pd_inner, target=label)
        w = self.modeltpl_cfg['loss_weight']
        # loss = loss_all+w[0]*loss_context+w[1]*loss_inner
        return {
            'loss_main': loss_all,
            'loss_context': w[0] * loss_context,
            'loss_inner': w[1] * loss_inner
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)

class ECD_MIRT(ECD):
    default_cfg = {
        "a_range": -1.0,  # disc range
        "emb_dim": 32,
        'loss_weight': [0, 4],
    }
    def build_model(self):
        super().build_model()
        self.emb_dim = self.modeltpl_cfg['emb_dim']
        self.theta = nn.Embedding(self.n_user, self.emb_dim)  # student ability
        self.a = nn.Embedding(self.n_item, self.emb_dim)  # exer discrimination
        self.b = nn.Embedding(self.n_item, 1)  # exer intercept term

    @staticmethod
    def irf(theta, a, b):
        return 1 / (1 + torch.exp(- torch.sum(torch.multiply(a, theta), axis=-1) + b))

    def forward(self, stu_id, exer_id):
        items_Q_mat = self.Q_mat[exer_id]
        stus_qqq = self.QQQ_list[stu_id]
        theta_inner = self.theta(stu_id)
        a = self.a(exer_id)
        b = self.b(exer_id).flatten()
        theta_context, dt_w = super().forward(stu_id, stus_qqq)
        theta_all = theta_context * dt_w + theta_inner * (1 - dt_w)
        if self.modeltpl_cfg['a_range'] is not None:
            a = self.modeltpl_cfg['a_range'] * torch.sigmoid(a)
        else:
            a = F.softplus(a) # 让区分度大于0，保持单调性假设
        if torch.max(theta_inner != theta_inner) or torch.max(a != a) or torch.max(b != b):  # pragma: no cover
            raise ValueError('ValueError:theta,a,b may contains nan!  The diff_range or a_range is too large.')
        pre_all = self.irf(theta_all, a, b)
        pre_context = self.irf(theta_context, a, b)
        pre_inner = self.irf(theta_inner, a, b)
        if self.training:
            return pre_all, pre_context, pre_inner
        else:
            return pre_all

    @torch.no_grad()
    def predict(self, stu_id, exer_id, **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd_all, pd_context, pd_inner = self(stu_id, exer_id)
        pd_all, pd_context, pd_inner = pd_all.flatten(), pd_context.flatten(), pd_inner.flatten()
        loss_all = F.binary_cross_entropy(input=pd_all, target=label)
        loss_context = F.binary_cross_entropy(input=pd_context, target=label)
        loss_inner = F.binary_cross_entropy(input=pd_inner, target=label)
        w = self.modeltpl_cfg['loss_weight']
        # loss = loss_all+w[0]*loss_context+w[1]*loss_inner
        return {
            'loss_main': loss_all,
            'loss_context': w[0] * loss_context,
            'loss_inner': w[1] * loss_inner
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)

class ECD_NCD(ECD):
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'disc_scale': 10,
        'loss_weight': [1, 1],
    }
    def build_model(self):
        super().build_model()
        self.student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.k_difficulty = nn.Embedding(self.n_item, self.n_cpt)
        self.e_difficulty = nn.Embedding(self.n_item, 1)
        self.pd_net = PosMLP(
            input_dim=self.n_cpt, output_dim=1, activation=self.modeltpl_cfg['activation'],
            dnn_units=self.modeltpl_cfg['dnn_units'], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )

    def forward(self, stu_id, exer_id):
        items_Q_mat = self.Q_mat[exer_id]
        stus_qqq = self.QQQ_list[stu_id]

        stu_emb = self.student_emb(stu_id)
        theta_inner = torch.sigmoid(stu_emb)
        k_difficulty = torch.sigmoid(self.k_difficulty(exer_id))
        e_difficulty = torch.sigmoid(self.e_difficulty(exer_id)) * self.modeltpl_cfg['disc_scale']
        # prednet
        input_knowledge_point = items_Q_mat
        theta_context, dt_w = super().forward(stu_id, stus_qqq)
        theta_all = theta_context * dt_w + theta_inner * (1 - dt_w)

        pre_all = self.pd_net(e_difficulty * (theta_all - k_difficulty) * input_knowledge_point).sigmoid()
        pre_context = self.pd_net(e_difficulty * (theta_context - k_difficulty) * input_knowledge_point).sigmoid()

        input_x = e_difficulty * (theta_inner - k_difficulty) * input_knowledge_point
        pre_inner = self.pd_net(input_x).sigmoid()
        if self.training:
            return pre_all, pre_context, pre_inner
        else:
            return pre_all

    @torch.no_grad()
    def predict(self, stu_id, exer_id,  **kwargs):
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }

    def get_main_loss(self, **kwargs):
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd_all, pd_context, pd_inner = self(stu_id, exer_id)
        pd_all, pd_context, pd_inner = pd_all.flatten(), pd_context.flatten(), pd_inner.flatten()
        loss_all = F.binary_cross_entropy(input=pd_all, target=label)
        loss_context = F.binary_cross_entropy(input=pd_context, target=label)
        loss_inner = F.binary_cross_entropy(input=pd_inner, target=label)
        w = self.modeltpl_cfg['loss_weight']
        # loss = loss_all+w[0]*loss_context+w[1]*loss_inner
        return {
            'loss_main': loss_all,
            'loss_context': w[0]*loss_context,
            'loss_inner': w[1]*loss_inner
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



from .base_mid2cache import BaseMid2Cache
import pandas as pd
from itertools import chain


class M2C_MergeDividedSplits(BaseMid2Cache):
    default_cfg = {}

    def process(self, **kwargs):
        df_train = kwargs['df_train']
        df_valid = kwargs['df_valid']
        df_test = kwargs['df_test']
        
        # 1. ensure the keys in df_train, df_valid, df_test is same
        assert df_train is not None and df_test is not None
        assert set(df_train.columns) == set(df_test.columns)
        # 2. 包容df_valid不存在的情况

        df = pd.concat((df_train, df_test), ignore_index=True)

        if df_valid is not None:
            assert set(df_train.columns) == set(df_valid.columns)
            df = pd.concat((df, df_valid), ignore_index=True)
        
        kwargs['df'] = df
        return kwargs

    def set_dt_info(self, dt_info, **kwargs):
        if 'stu_id:token' in kwargs['df'].columns:
            dt_info['stu_count'] = int(kwargs['df']['stu_id:token'].max() + 1)
        if 'exer_id:token' in kwargs['df'].columns:
            dt_info['exer_count'] = int(kwargs['df']['exer_id:token'].max() + 1)
        if kwargs.get('df_exer', None) is not None:
            if 'cpt_seq:token_seq' in kwargs['df_exer']:
                dt_info['cpt_count'] = len(set(list(chain(*kwargs['df_exer']['cpt_seq:token_seq'].to_list()))))



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from .callback import Callback

class TensorboardCallback(Callback):
    def __init__(self, log_dir, comment=''):
        super(TensorboardCallback, self).__init__()
        from torch.utils.tensorboard import SummaryWriter
        self.writer = SummaryWriter(log_dir=log_dir, comment=comment)

    def on_epoch_end(self, epoch, logs=None, **kwargs):
        for name, value in logs.items():
            self.writer.add_scalar(tag=name, scalar_value=value, global_step=epoch)



from .base_evaltpl import BaseEvalTPL
import numpy as np
from edustudio.utils.common import tensor2npy
from sklearn.metrics import mean_squared_error, roc_auc_score, accuracy_score, f1_score, label_ranking_loss, coverage_error


class PredictionEvalTPL(BaseEvalTPL):
    """Student Performance Prediction Evaluation
    """
    default_cfg = {
        'use_metrics': ['auc', 'acc', 'rmse']
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    
    def eval(self, y_pd, y_gt, **kwargs):
        if not isinstance(y_pd, np.ndarray): y_pd = tensor2npy(y_pd)
        if not isinstance(y_gt, np.ndarray): y_gt = tensor2npy(y_gt)
        metric_result = {}
        ignore_metrics = kwargs.get('ignore_metrics', {})
        for metric_name in self.evaltpl_cfg[self.__class__.__name__]['use_metrics']:
            if metric_name not in ignore_metrics:
                metric_result[metric_name] = self._get_metrics(metric_name)(y_gt, y_pd)
        return metric_result
        
    def _get_metrics(self, metric):
        if metric == "auc":
            return roc_auc_score
        elif metric == "mse":
            return mean_squared_error
        elif metric == 'rmse':
            return lambda y_gt, y_pd: mean_squared_error(y_gt, y_pd) ** 0.5
        elif metric == "acc":
            return lambda y_gt, y_pd: accuracy_score(y_gt, np.where(y_pd >= 0.5, 1, 0))
        elif metric == "f1_macro":
            return lambda y_gt, y_pd: f1_score(y_gt, y_pd, average='macro')
        elif metric == "f1_micro":
            return lambda y_gt, y_pd: f1_score(y_gt, y_pd, average='micro')
        elif metric == "ranking_loss":
            return lambda y_gt, y_pd: label_ranking_loss(y_gt, y_pd)
        elif metric == 'coverage_error':
            return lambda y_gt, y_pd: coverage_error(y_gt, y_pd)
        elif metric == 'samples_auc':
            return lambda y_gt, y_pd: roc_auc_score(y_gt, y_pd, average='samples')
        else:
            raise NotImplementedError




from .dkt import DKT
import torch.nn as nn
import torch
import torch.nn.functional as F


class DKT_plus(DKT):
    default_cfg = {
        'lambda_r': 0.01,
        'lambda_w1': 0.003,
        'lambda_w2': 3.0,
        'reg_all_KCs': True
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def get_main_loss(self, **kwargs):
        q = kwargs['exer_seq'][:, :-1].unsqueeze(dim=-1)
        q_shft = kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1)

        y = self(**kwargs)
        pred = y[:, :-1].gather(index=q, dim=2).squeeze(dim=-1)
        pred_shft = y[:, :-1].gather(index=q_shft, dim=2).squeeze(dim=-1)

        y_curr = pred[kwargs['mask_seq'][:, :-1] == 1]
        y_next = pred_shft[kwargs['mask_seq'][:, 1:] == 1]

        gt_curr =  kwargs['label_seq'][:, :-1][kwargs['mask_seq'][:, :-1] == 1]
        gt_next = kwargs['label_seq'][:, 1:][kwargs['mask_seq'][:, 1:] == 1]

        loss_main = F.binary_cross_entropy(input=y_next, target=gt_next)
        loss_r = self.modeltpl_cfg['lambda_r'] * F.binary_cross_entropy(input=y_curr, target=gt_curr)

        if self.modeltpl_cfg['reg_all_KCs']:
            diff = y[:, 1:] - y[:, :-1]
        else:
            diff = (pred_shft - pred)[kwargs['mask_seq'][:, 1:] == 1]
        loss_w1 = torch.norm(diff, 1) / len(diff)
        loss_w1 = self.modeltpl_cfg['lambda_w1'] * loss_w1 / self.n_item
        loss_w2 = torch.norm(diff, 2) / len(diff)
        loss_w2 = self.modeltpl_cfg['lambda_w2'] * loss_w2 / self.n_item

        return {
            'loss_main': loss_main,
            'loss_r': loss_r,
            'loss_w1': loss_w1,
            'loss_w2': loss_w2
        }



import fire
from edustudio.atom_op.raw2mid import _cli_api_dict_ as raw2mid
from edustudio.utils.common import Logger

def entrypoint():
    Logger().get_std_logger()
    fire.Fire(
        {"r2m": raw2mid}
    )



from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.autograd import Variable

class KQN(GDBaseModel):
    default_cfg = {
        'emb_size': 128,
        'rnn_hidden_size': 128,
        'mlp_hidden_size': 128,
        'n_rnn_layers': 1,
        'rnn_or_lstm': 'lstm',
        'dropout': 0.2
    }

    def __init__(self, cfg):
        super().__init__(cfg)
    
    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
    
    def build_model(self):
        # helper variable for making a one-hot vector for rnn input
        self.inter_emb = torch.eye(self.n_item * 2).to(self.device)
        # helper variable for making a one-hot vector for skills
        self.cpt_emb = torch.eye(self.n_item).to(self.device)

        if self.modeltpl_cfg['rnn_or_lstm'] == 'gru':
            self.seq_model = nn.GRU(
                2 * self.n_item, self.modeltpl_cfg['rnn_hidden_size'], 
                self.modeltpl_cfg['n_rnn_layers'], batch_first=True
            )
        else:
            self.seq_model = nn.LSTM(
                2 * self.n_item, self.modeltpl_cfg['rnn_hidden_size'], 
                self.modeltpl_cfg['n_rnn_layers'], batch_first=True
            )
        
        self.skill_encoder = nn.Sequential(
            nn.Linear(self.n_item, self.modeltpl_cfg['mlp_hidden_size']),
            nn.ReLU(),
            nn.Linear(self.modeltpl_cfg['mlp_hidden_size'], self.modeltpl_cfg['emb_size']),
            nn.ReLU()
        )
        self.fc_layer = nn.Linear(self.modeltpl_cfg['rnn_hidden_size'], self.modeltpl_cfg['emb_size'])
        self.drop_layer = nn.Dropout(self.modeltpl_cfg['dropout'])
        self.sigmoid = nn.Sigmoid()

    def forward(self, exer_seq, label_seq, **kwargs):
        in_data = self.inter_emb[(exer_seq[:,:-1] + label_seq[:,:-1] * self.n_item).long()]
        next_skills = self.cpt_emb[(exer_seq[:,1:]).long()]

        encoded_knowledge = self.encode_knowledge(in_data.to(self.traintpl_cfg['device']))
        encoded_skills = self.encode_skills(next_skills.to(self.traintpl_cfg['device']))
        encoded_knowledge = self.drop_layer(encoded_knowledge)

        y_pd = torch.sum(encoded_knowledge * encoded_skills, dim=2) # (batch_size, max_seq_len)
        y_pd = self.sigmoid(y_pd)

        return y_pd

    def encode_knowledge(self, in_data):
        batch_size = in_data.size(0)
        self.hidden = self.init_hidden(batch_size)
        
        # rnn_input = pack_padded_sequence(in_data, seq_len, batch_first=True)
        rnn_output, _ = self.seq_model(in_data, self.hidden)
        # rnn_output, _ = pad_packed_sequence(rnn_output, batch_first=True) # (batch_size, max_seq_len, n_rnn_hidden)
        encoded_knowledge = self.fc_layer(rnn_output) # (batch_size, max_seq_len, n_hidden)

        return encoded_knowledge
    
    def encode_skills(self, next_skills):
        encoded_skills = self.skill_encoder(next_skills) # (batch_size, max_seq_len, n_hidden)
        encoded_skills = F.normalize(encoded_skills, p=2, dim=2) # L2-normalize

        return encoded_skills

    def init_hidden(self, batch_size: int):
        weight = next(self.parameters()).data
        if self.modeltpl_cfg['rnn_or_lstm'] == 'lstm':
            return (Variable(weight.new(self.modeltpl_cfg['n_rnn_layers'], batch_size, self.modeltpl_cfg['rnn_hidden_size']).zero_()),
                    Variable(weight.new(self.modeltpl_cfg['n_rnn_layers'], batch_size, self.modeltpl_cfg['rnn_hidden_size']).zero_()))
        else:
            return Variable(weight.new(self.modeltpl_cfg['n_rnn_layers'], batch_size, self.modeltpl_cfg['rnn_hidden_size']).zero_())

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        # y_pd = y_pd[:-1].gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=1
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        # y_pd = y_pd.gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

from typing import List
from .callback import Callback
import numpy as np


class Metric(object):
    def __init__(self, name, type_):
        assert type_ in ['max', 'min']
        self.name = name
        self.type_ = type_
        self.best_epoch = 1
        if self.type_ == 'max':
            self.best_value = -np.inf
        else:
            self.best_value = np.inf

    def better_than(self, value):
        return value < self.best_value if self.type_ == 'max' else value > self.best_value

    def update(self, epoch, value):
        self.best_epoch = epoch
        self.best_value = value


class EarlyStopping(Callback):
    def __init__(self, metric_list:List[list], num_stop_rounds: int = 20, start_round=1):
        """_summary_

        Args:
            metric_list (List[list]): [['rmse', 'min'],['ndcg', 'max']]
            num_stop_rounds (int, optional): all metrics have no improvement in latest num_stop_rounds, suggest to stop training. Defaults to 20.
            start_round (int, optional): start detecting from epoch start_round, . Defaults to 1.
        """
        super().__init__()
        assert num_stop_rounds >= 1
        assert start_round >= 1
        self.start_round = start_round
        self.num_stop_round = num_stop_rounds
        self.stop_training = False
        self.metric_list = [Metric(name=metric_name, type_=metric_type) for metric_name, metric_type in metric_list]

    def on_train_begin(self, logs=None, **kwargs):
        super().on_train_begin()
        self.model.share_callback_dict['stop_training'] = False

    def on_epoch_end(self, epoch: int, logs: dict = None, **kwargs):
        flag = True
        for metric in self.metric_list:
            if not metric.better_than(logs[metric.name]):
                metric.update(epoch=epoch, value=logs[metric.name])
        
            if self.start_round <= epoch:
                if epoch - metric.best_epoch < self.num_stop_round:
                    flag &= False

        if flag is True:
            self.logger.info("Suggest to stop training now")
            self.stop_training = True
            self.model.share_callback_dict['stop_training'] = True



import numpy as np
from sklearn.model_selection import StratifiedKFold, KFold


class SpliterUtil:
    @staticmethod
    def divide_data_df_one_fold(df, divide_scale_list, seed, label_field=None, shuffle=True):
        n_split_1, n_split_2 = None, None
        if len(divide_scale_list) == 2:
            n_split_1 = np.sum(divide_scale_list) / divide_scale_list[-1]
        elif len(divide_scale_list) == 3:
            n_split_1 = np.sum(divide_scale_list) / divide_scale_list[-1]
            n_split_2 = np.sum(divide_scale_list[0:2]) / divide_scale_list[1]
        else:
            raise ValueError(f"unsupported divide_scale_list={divide_scale_list}")

        if label_field:
            skf = StratifiedKFold(n_splits=int(n_split_1),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(df, df[label_field])
        else:
            skf = KFold(n_splits=int(n_split_1),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(df)

        train_df, val_df, test_df = None, None, None
        for train_index, test_index in splits:
            train_df_tmp = df.iloc[train_index].reset_index(drop=True)
            test_df = df.iloc[test_index].reset_index(drop=True)
            break

        if n_split_2 is not None:
            if label_field:
                skf = StratifiedKFold(n_splits=int(n_split_2),
                                    shuffle=True, random_state=seed)
                splits = skf.split(train_df_tmp, train_df_tmp[label_field])
            else:
                skf = KFold(n_splits=int(n_split_2),
                                    shuffle=True, random_state=seed)
                splits = skf.split(train_df_tmp)
            for train_index, val_index in splits:
                train_df = train_df_tmp.iloc[train_index].reset_index(
                    drop=True)
                val_df = train_df_tmp.iloc[val_index].reset_index(drop=True)
                break
        else:
            train_df = train_df_tmp
        return train_df, val_df, test_df
    
    @staticmethod
    def divide_data_df_multi_folds(df, n_folds, seed, label_field=None, shuffle=True):
        if label_field:
            skf = StratifiedKFold(n_splits=int(n_folds),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(df, df[label_field])
        else:
            skf = KFold(n_splits=int(n_folds),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(df)

        train_df_list, test_df_list = [], []
        for train_index, test_index in splits:
            train_df = df.iloc[train_index].reset_index(drop=True)
            test_df = df.iloc[test_index].reset_index(drop=True)
            train_df_list.append(train_df)
            test_df_list.append(test_df)

        return train_df_list, test_df_list

    @staticmethod
    def divide_data_dict(dic, divide_scale_list, seed, label_field=None, shuffle=True):
        keys = list(dic.keys())
        n_split_1, n_split_2 = None, None
        if len(divide_scale_list) == 2:
            n_split_1 = np.sum(divide_scale_list) / divide_scale_list[-1]
        elif len(divide_scale_list) == 3:
            n_split_1 = np.sum(divide_scale_list) / divide_scale_list[-1]
            n_split_2 = np.sum(divide_scale_list[0:2]) / divide_scale_list[1]
        else:
            raise ValueError(f"unsupported divide_scale_list={divide_scale_list}")

        if label_field:
            skf = StratifiedKFold(n_splits=int(n_split_1),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(dic[keys[0]], dic[label_field])
        else:
            skf = KFold(n_splits=int(n_split_1),
                                shuffle=shuffle, random_state=seed)
            splits = skf.split(dic[keys[0]])

        train_dic, val_dic, test_dic = None, None, None
        for train_index, test_index in splits:
            train_dic_tmp = {dic[k][train_index] for k in keys}
            test_dic = {dic[k][test_index] for k in keys}
            break

        if n_split_2 is not None:
            if label_field:
                skf = StratifiedKFold(n_splits=int(n_split_2),
                                    shuffle=True, random_state=seed)
                splits = skf.split(train_dic_tmp[keys[0]], train_dic_tmp[label_field])
            else:
                skf = KFold(n_splits=int(n_split_2), shuffle=True, random_state=seed)
                splits = skf.split(train_dic_tmp)
            for train_index, val_index in splits:
                train_dic = {train_dic_tmp[k][train_index] for k in keys}
                val_dic = {train_dic_tmp[k][val_index] for k in keys}
                break
        else:
            train_dic = train_dic_tmp
        return train_dic, val_dic, test_dic



import logging
from edustudio.utils.common import UnifyConfig
from torch.utils.data import Dataset
import copy
from ..utils.common import BigfileDownloader, DecompressionUtil
import yaml
import re
import os
import requests


class BaseDataTPL(Dataset):
    """The basic data protocol for implementing a data template
    """
    default_cfg = {'seed': 2023}

    def __init__(self, cfg):
        self.cfg: UnifyConfig = cfg
        self.datatpl_cfg: UnifyConfig = cfg.datatpl_cfg
        self.evaltpl_cfg: UnifyConfig = cfg.evaltpl_cfg
        self.traintpl_cfg: UnifyConfig = cfg.traintpl_cfg
        self.frame_cfg: UnifyConfig = cfg.frame_cfg
        self.modeltpl_cfg: UnifyConfig = cfg.modeltpl_cfg
        self.logger: logging.Logger = logging.getLogger("edustudio")
        self._check_params()

    @classmethod
    def get_default_cfg(cls, **kwargs):
        """Get the final default config object

        Returns:
            UnifyConfig: the final default config object
        """
        cfg = UnifyConfig()
        for _cls in cls.__mro__:
            if not hasattr(_cls, 'default_cfg'):
                break
            cfg.update(_cls.default_cfg, update_unknown_key_only=True)
        return cfg

    @classmethod
    def from_cfg(cls, cfg):
        """Instantiate data template

        Args:
            cfg (UnifyConfig): the global config object

        Returns:
            BaseDataTPL: instance of data template
        """
        return cls(cfg)
    
    def get_extra_data(self, **kwargs):
        """an interface to construct extra data except the data from forward API
        """
        return {}

    def _check_params(self):
        """check validation of default config
        """
        pass

    def _copy(self):
        """copy current instance
        """
        obj = copy.copy(self)
        return obj

    @classmethod
    def download_dataset(cls, cfg):
        """Download Dataset from the Internet

        Args:
            cfg (UnifyConfig):the global config object
        """
        dt_name = cfg.dataset
        cfg.logger.warning(f"Can't find dataset files of {dt_name} in local disk")
        
        fph = cfg.frame_cfg['DT_INFO_FILE_PATH']
        dataset_info = cls.read_yml_file(fph)
        dataset_info_from_cfg: dict = cfg['frame_cfg']['DT_INFO_DICT']
        dataset_info.update(dataset_info_from_cfg)

        if dt_name not in dataset_info:
            cfg.logger.info(f"Prepare download external datasets.yaml to find dataset:{dt_name}")
            url = "https://huggingface.co/datasets/lmcRS/edustudio-datasets/raw/main/datasets.yaml"
            cfg.logger.info(f"Eexternal datasets.yaml url: {url}")
            resp = requests.get(url)
            dataset_info_external = yaml.load(resp.text, Loader=cls._build_yaml_loader())
            if dt_name not in dataset_info_external:
                raise Exception("Can't find dataset files from local disk and online")
            else:
                dataset_info.update(dataset_info_external)

        cfg.logger.info(f"Prepare to download {dt_name} dataset from online")
        cfg.logger.info(f"Download_url: {dataset_info[dt_name]['middata_url']}")

        if not os.path.exists(cfg.frame_cfg.data_folder_path):
            os.makedirs(cfg.frame_cfg.data_folder_path)
        fph_tmp = f"{cfg.frame_cfg.data_folder_path}/{dt_name}-middata.zip.tmp"
        fph_zip = f"{cfg.frame_cfg.data_folder_path}/{dt_name}-middata.zip"
        if not os.path.exists(fph_zip):
            if os.path.exists(fph_tmp):
                os.remove(fph_tmp)
            BigfileDownloader.download(
                url=dataset_info[dt_name]['middata_url'], title=f"{dt_name} downloading...", 
                filepath=fph_tmp
            )
            os.rename(fph_tmp, fph_zip)
        else:
            cfg.logger.info(f"Find a zip file of {dt_name} at {fph_zip}, skip downloading process")

        DecompressionUtil.unzip_file(
            zip_src=fph_zip, dst_dir=cfg.frame_cfg.data_folder_path
        )
        
    @classmethod
    def read_yml_file(cls, filepath: str):
        with open(filepath, 'r', encoding='utf-8') as f:
            config = yaml.load(f, Loader=cls._build_yaml_loader())
        return config

    @staticmethod
    def _build_yaml_loader():
        loader = yaml.FullLoader
        loader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(
                u'''^(?:
             [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$''', re.X
            ), list(u'-+0123456789.')
        )
        return loader



r"""
LPKT
##########################################

Reference:
    Shuanghong Sheng et al. "Learning Process-consistent Knowledge Tracing" in KDD 2021.

Reference Code:
    https://github.com/bigdata-ustc/EduKTM/tree/main/EduKTM/LPKT

"""

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F


class LPKT(GDBaseModel):
    default_cfg = {
        'd_a': 50,
        'd_k': 128,
        'd_e': 128, 
        'q_gamma': 0.03,
        'drop_rate': 0.2,
        'param_init_type': 'xavier_uniform'
    }

    def build_cfg(self):
        self.n_stu = self.datatpl_cfg['dt_info']['stu_count']
        self.n_exer = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']
        self.n_at = self.datatpl_cfg['dt_info']['answer_time_count'] # answer time
        self.n_it = self.datatpl_cfg['dt_info']['interval_time_count'] # interval time
        self.d_k = self.modeltpl_cfg['d_k']
        self.d_e = self.modeltpl_cfg['d_e']
        self.d_a = self.modeltpl_cfg['d_a']

    def add_extra_data(self, **kwargs):
        Q_mat = kwargs['Q_mat']
        Q_mat[Q_mat == 0] = self.modeltpl_cfg['q_gamma']
        self.q_matrix = Q_mat.to(self.device).float()

    def build_model(self):
        d_k = self.modeltpl_cfg['d_k']
        d_e = self.modeltpl_cfg['d_e']
        d_a = self.modeltpl_cfg['d_a']

        self.at_embed = nn.Embedding(self.n_at, d_k)
        self.it_embed = nn.Embedding(self.n_it, d_k)
        self.exer_embed = nn.Embedding(self.n_exer, d_e)

        self.linear_1 = nn.Linear(d_a + d_e + d_k, d_k)
        self.linear_2 = nn.Linear(4 * d_k, d_k)
        self.linear_3 = nn.Linear(4 * d_k, d_k)
        self.linear_4 = nn.Linear(3 * d_k, d_k)
        self.linear_5 = nn.Linear(2 * d_k, d_k)

        self.dropout = nn.Dropout(self.modeltpl_cfg['drop_rate'])

    def forward(self, exer_seq, answer_time_seq, interval_time_seq, **kwargs):
        a_data = exer_seq
        batch_size, seq_len = exer_seq.shape
        e_embed_data = self.exer_embed(exer_seq)
        at_embed_data = self.at_embed(answer_time_seq)
        it_embed_data = self.it_embed(interval_time_seq)
        a_data = a_data.view(-1, 1).repeat(1, self.d_a).view(batch_size, -1, self.d_a)
        h_pre = nn.init.xavier_uniform_(torch.zeros(self.n_cpt, self.d_k)).repeat(batch_size, 1, 1).to(self.device)
        h_tilde_pre = None
        all_learning = self.linear_1(torch.cat((e_embed_data, at_embed_data, a_data), 2))
        learning_pre = torch.zeros(batch_size, self.d_k).to(self.device)

        pred = torch.zeros(batch_size, seq_len).to(self.device)
        for t in range(0, seq_len - 1):
            e = exer_seq[:, t]
            # q_e: (bs, 1, n_skill)
            q_e = self.q_matrix[e].view(batch_size, 1, -1)
            it = it_embed_data[:, t]

            # Learning Module
            if h_tilde_pre is None:
                h_tilde_pre = q_e.bmm(h_pre).view(batch_size, self.d_k)
            learning = all_learning[:, t]
            learning_gain = self.linear_2(torch.cat((learning_pre, it, learning, h_tilde_pre), 1))
            learning_gain = learning_gain.tanh()
            gamma_l = self.linear_3(torch.cat((learning_pre, it, learning, h_tilde_pre), 1))
            gamma_l = gamma_l.sigmoid()
            LG = gamma_l * ((learning_gain + 1) / 2)
            LG_tilde = self.dropout(q_e.transpose(1, 2).bmm(LG.view(batch_size, 1, -1)))

            # Forgetting Module
            # h_pre: (bs, n_skill, d_k)
            # LG: (bs, d_k)
            # it: (bs, d_k)
            n_skill = LG_tilde.size(1)
            gamma_f = self.linear_4(torch.cat((
                h_pre,
                LG.repeat(1, n_skill).view(batch_size, -1, self.d_k),
                it.repeat(1, n_skill).view(batch_size, -1, self.d_k)
            ), 2)).sigmoid()
            h = LG_tilde + gamma_f * h_pre

            # Predicting Module
            h_tilde = self.q_matrix[exer_seq[:, t + 1]].view(batch_size, 1, -1).bmm(h).view(batch_size, self.d_k)
            y = self.linear_5(torch.cat((e_embed_data[:, t + 1], h_tilde), 1)).sigmoid().sum(1) / self.d_k
            pred[:, t + 1] = y

            # prepare for next prediction
            learning_pre = learning
            h_pre = h
            h_tilde_pre = h_tilde
        return pred 

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)[:, 0:-1]
        # y_pd = y_pd[:, :-1].gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)[:,0:-1] # 可能与原文不符合，原文是[:,1:]
        # y_pd = y_pd[:, :-1].gather(
        #     index=kwargs['exer_seq'][:, 1:].unsqueeze(dim=-1), dim=2
        # ).squeeze(dim=-1)
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)



from ..common import EduDataTPL
import torch
import pandas as pd
import os

class CNCDQDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 'M2C_RandomDataSplit4CD', 'M2C_GenQMat'],
    }

    @property
    def common_str2df(self):
        return {
            "df": self.df, "df_train": self.df_train, "df_valid": self.df_valid,
            "df_test": self.df_test, "dt_info": self.datatpl_cfg['dt_info'], 
            "df_stu": self.df_stu, "df_exer": self.df_exer,
            "df_questionnaire": self.df_questionnaire
        }

    @classmethod
    def load_data(cls, cfg): # 只在middata存在时调用
        kwargs = super().load_data(cfg)
        new_kwargs = cls.load_data_from_questionnaire(cfg)
        for df in new_kwargs.values(): 
            if df is not None:
                cls._preprocess_feat(df) # 类型转换
        kwargs.update(new_kwargs)
        return kwargs

    @classmethod
    def load_data_from_questionnaire(cls, cfg):
        file_path = f'{cfg.frame_cfg.data_folder_path}/middata/{cfg.dataset}.questionnaire.csv'
        df_questionnaire = None
        if os.path.exists(file_path):
            sep = cfg.datatpl_cfg['seperator']
            df_questionnaire = pd.read_csv(file_path, sep=sep, encoding='utf-8', usecols=['cpt_head:token', 'cpt_tail:token'])
        return {"df_questionnaire": df_questionnaire}



from torch.utils.data import DataLoader
import random
from ..common import BaseProxyDataTPL
import numpy

class PairDataLoader(DataLoader):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    
    def __iter__(self):
        iter_obj = super().__iter__()
        self.dataset.sample()
        return iter_obj



class IRRDataTPL(BaseProxyDataTPL):
    default_cfg = {
        'backbone_datatpl_cls': 'CDInterDataTPL',
        'num_observed': 10,
        'num_unobserved': 0
    }

    def sample(self):
        self.stu_count = self.datatpl_cfg['dt_info']['stu_count']
        self.num_observed = self.datatpl_cfg['num_observed']
        self.num_unobserved = self.datatpl_cfg['num_unobserved']
        stu_id = self.dict_main['stu_id'].numpy()
        exer_id = self.dict_main['exer_id'].numpy()
        label = self.dict_main['label'].numpy()
        students_with_correct_answers = {}
        students_with_wrong_answers = {}

        for i in range(stu_id.shape[0]):
            if label[i] == 0:
                students_with_wrong_answers.setdefault(exer_id[i], []).append(stu_id[i])
            else:
                students_with_correct_answers.setdefault(exer_id[i], []).append(stu_id[i])

        self.dict_main['pair_exer'] = []
        self.dict_main['pair_pos_stu'] = []
        self.dict_main['pair_neg_stu'] = []

        for i in range(stu_id.shape[0]):
            self.pair_exer = exer_id[i]

            if label[i] == 0:
                self.pair_neg_stu = stu_id[i]
                do_right_stu_exer = students_with_correct_answers.get(self.pair_exer, [])
                do_right_stu_exer_length = len(do_right_stu_exer)

                if do_right_stu_exer_length < self.num_observed:
                    for j in range(do_right_stu_exer_length):
                        self.add_pair(self.pair_exer, do_right_stu_exer[j], self.pair_neg_stu)
                else:
                    pos_stu_list = random.sample(do_right_stu_exer, k=self.num_observed)
                    for j in range(self.num_observed):
                        self.add_pair(self.pair_exer, pos_stu_list[j], self.pair_neg_stu)

                unobserved = []
                do_wrong_stu_exer = students_with_wrong_answers.get(self.pair_exer, [])

                while len(unobserved) < self.num_unobserved and len(unobserved) < do_right_stu_exer_length:
                    self.pair_pos_stu = numpy.int64(random.randint(0, self.stu_count - 1))

                    while (self.pair_pos_stu in do_right_stu_exer) or (self.pair_pos_stu in do_wrong_stu_exer) or (self.pair_pos_stu in unobserved):
                        self.pair_pos_stu = numpy.int64(random.randint(0, self.stu_count - 1))

                    unobserved.append(self.pair_pos_stu)

                for j in range(len(unobserved)):
                    self.add_pair(self.pair_exer, unobserved[j], self.pair_neg_stu)
            else:
                self.pair_pos_stu = stu_id[i]
                do_wrong_stu_exer = students_with_wrong_answers.get(self.pair_exer, [])
                do_wrong_stu_exer_length = len(do_wrong_stu_exer)

                if do_wrong_stu_exer_length < self.num_observed:
                    for j in range(do_wrong_stu_exer_length):
                        self.add_pair(self.pair_exer, self.pair_pos_stu, do_wrong_stu_exer[j])
                else:
                    neg_stu_list = random.sample(do_wrong_stu_exer, k=self.num_observed)
                    for j in range(self.num_observed):
                        self.add_pair(self.pair_exer, self.pair_pos_stu, neg_stu_list[j])

                unobserved = []
                do_right_stu_exer = students_with_correct_answers.get(self.pair_exer, [])

                while len(unobserved) < self.num_unobserved and len(unobserved) < do_wrong_stu_exer_length:
                    self.pair_neg_stu = numpy.int64(random.randint(0, self.stu_count - 1))

                    while (self.pair_neg_stu in do_right_stu_exer) or (self.pair_neg_stu in do_wrong_stu_exer) or (self.pair_neg_stu in unobserved):
                        self.pair_neg_stu = numpy.int64(random.randint(0, self.stu_count - 1))

                    unobserved.append(self.pair_neg_stu)

                for j in range(len(unobserved)):
                    self.add_pair(self.pair_exer, self.pair_pos_stu, unobserved[j])

    def add_pair(self, exer_id, pos_stu, neg_stu):
        self.dict_main['pair_exer'].append(exer_id)
        self.dict_main['pair_pos_stu'].append(pos_stu)
        self.dict_main['pair_neg_stu'].append(neg_stu)


    def build_dataloaders(self):
        batch_size = self.traintpl_cfg['batch_size']
        num_workers = self.traintpl_cfg['num_workers']
        eval_batch_size = self.traintpl_cfg['eval_batch_size']
        train_dt_list, valid_dt_list, test_dt_list = self.build_datasets()
        train_loader_list, valid_loader_list, test_loader_list = [], [], []

        for fid in range(self.datatpl_cfg['n_folds']):
            train_loader = PairDataLoader(dataset=train_dt_list[fid], shuffle=True, batch_size=batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
            train_loader_list.append(train_loader)
            if self.hasValidDataset:
                valid_loader = PairDataLoader(dataset=valid_dt_list[fid], shuffle=False, batch_size=eval_batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
                valid_loader_list.append(valid_loader)
            test_loader = PairDataLoader(dataset=test_dt_list[fid], shuffle=False, batch_size=eval_batch_size, num_workers=num_workers, collate_fn=self.collate_fn)
            test_loader_list.append(test_loader)
        
        return train_loader_list, valid_loader_list, test_loader_list



import math

from ..gd_basemodel import GDBaseModel
import torch.nn as nn
import torch
import torch.nn.functional as F
import numpy as np
from torch.nn.init import xavier_uniform_, constant_

class SimpleKT(GDBaseModel):
    default_cfg = {
        'l2': 1e-5,
        'kq_same': 1,
        'dropout_rate': 0.05,
        'separate_qa': False,
        'd_model': 256,
        'n_blocks': 1,
        'final_fc_dim': 512,
        'n_heads': 8,
        'd_ff': 2048,
    }

    def __init__(self, cfg):
        super().__init__(cfg)

    def _init_params(self):
        super()._init_params()
        self.reset()

    def reset(self):
        for p in self.parameters():
            if p.size(0) == self.n_pid + 1 and self.n_pid > 0:
                constant_(p, 0.)

    def build_cfg(self):
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_question = self.datatpl_cfg['dt_info']['cpt_count']
        self.dropout = self.modeltpl_cfg['dropout_rate']
        self.kq_same = self.modeltpl_cfg['kq_same']
        self.l2 = self.modeltpl_cfg['l2']
        self.separate_qa = self.modeltpl_cfg['separate_qa']
        self.d_model = self.modeltpl_cfg['d_model']
        self.n_blocks = self.modeltpl_cfg['n_blocks']
        self.final_fc_dim = self.modeltpl_cfg['final_fc_dim']
        self.n_heads = self.modeltpl_cfg['n_heads']
        self.d_ff = self.modeltpl_cfg['d_ff']
        self.seq_len = self.datatpl_cfg['dt_info']['real_window_size']
        self.device = self.traintpl_cfg['device']

    def build_model(self):
        embed_l = self.d_model
        self.n_pid = self.n_item
        if self.n_pid > 0:
            self.difficult_param = nn.Embedding(self.n_pid + 1, 1)
            self.q_embed_diff = nn.Embedding(self.n_question + 1, embed_l)
            self.qa_embed_diff = nn.Embedding(2 * self.n_question + 1, embed_l)
        # n_question+1 ,d_model
        self.q_embed = nn.Embedding(self.n_question, embed_l)
        if self.separate_qa:
            self.qa_embed = nn.Embedding(2 * self.n_question + 1, embed_l)
        else:
            self.qa_embed = nn.Embedding(2, embed_l)
        # Architecture Object. It contains stack of attention block
        self.model = Architecture(n_blocks=self.n_blocks,
                                  n_heads=self.n_heads, dropout=self.dropout,
                                  d_model=self.d_model,
                                  d_ff=self.d_ff, kq_same=self.kq_same, seq_len=self.seq_len, device=self.device)

        self.out = nn.Sequential(
            nn.Linear(self.d_model + embed_l, self.final_fc_dim),
            nn.ReLU(),
            nn.Dropout(self.dropout),
            nn.Linear(self.final_fc_dim, 256),
            nn.ReLU(),
            nn.Dropout(self.dropout),
            nn.Linear(256, 1)
        )

    def forward(self, exer_seq, label_seq, cpt_unfold_seq, **kwargs):
        # Batch First
        # q_embed_data = self.q_embed(q_data)
        q_embed_data = self.q_embed(cpt_unfold_seq)
        if self.separate_qa:
            # qa_embed_data = self.qa_embed(qa_data)
            qa_data = cpt_unfold_seq + label_seq.long() * self.n_question
            qa_embed_data = self.qa_embed(qa_data)
        else:
            qa_data = label_seq.long()
            qa_embed_data = self.qa_embed(qa_data) + q_embed_data

        if self.n_pid > 0:
            q_embed_diff_data = self.q_embed_diff(cpt_unfold_seq)
            pid_embed_data = self.difficult_param(exer_seq)
            q_embed_data = q_embed_data + pid_embed_data * q_embed_diff_data
            qa_embed_diff_data = self.qa_embed_diff(qa_data)
            if self.separate_qa:
                qa_embed_data = qa_embed_data + pid_embed_data * qa_embed_diff_data
            else:
                qa_embed_data = qa_embed_data + pid_embed_data * (qa_embed_diff_data + q_embed_diff_data)



        d_output = self.model(q_embed_data, qa_embed_data)

        concat_q = torch.cat([d_output, q_embed_data], dim=-1)
        output = self.out(concat_q)
        pre = torch.sigmoid(output).squeeze(-1)
        return pre
        # if self.training:
        #     return pre, c_reg_loss
        # else:
        #     return pre

    @torch.no_grad()
    def predict(self, **kwargs):
        y_pd = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = None
        if kwargs.get('label_seq', None) is not None:
            y_gt = kwargs['label_seq'][:, 1:]
            y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        return {
            'y_pd': y_pd,
            'y_gt': y_gt
        }

    def get_main_loss(self, **kwargs):
        y_pd = self(**kwargs)
        # y_pd, c_reg_loss = self(**kwargs)
        y_pd = y_pd[:, :-1]
        y_pd = y_pd[kwargs['mask_seq'][:, 1:] == 1]
        y_gt = kwargs['label_seq'][:, 1:]
        y_gt = y_gt[kwargs['mask_seq'][:, 1:] == 1]
        loss = F.binary_cross_entropy(
            input=y_pd, target=y_gt
        )
        return {
            'loss_main': loss,
            # 'loss_reg': c_reg_loss
        }

    def get_loss_dict(self, **kwargs):
        return self.get_main_loss(**kwargs)


class Architecture(nn.Module):
    def __init__(self, n_blocks, d_model,
                 d_ff, n_heads, dropout, kq_same, seq_len, device):
        super().__init__()
        """
            n_block : number of stacked blocks in the attention
            d_model : dimension of attention input/output
            d_feature : dimension of input in each of the multi-head attention part.
            n_head : number of heads. n_heads*d_feature = d_model
        """
        self.d_model = d_model
        self.device = device


        self.blocks_2 = nn.ModuleList([
                TransformerLayer(d_model=d_model, d_feature=d_model // n_heads,
                                 d_ff=d_ff, dropout=dropout, n_heads=n_heads, kq_same=kq_same, device = device)
                for _ in range(n_blocks)
            ])
        self.position_emb = CosinePositionalEmbedding(d_model=self.d_model, max_len=seq_len)

    def forward(self, q_embed_data, qa_embed_data):
        # target shape  bs, seqlen
        seqlen, batch_size = q_embed_data.size(1), q_embed_data.size(0)

        q_posemb = self.position_emb(q_embed_data)
        q_embed_data = q_embed_data + q_posemb
        qa_posemb = self.position_emb(qa_embed_data)
        qa_embed_data = qa_embed_data + qa_posemb

        qa_pos_embed = qa_embed_data
        q_pos_embed = q_embed_data

        y = qa_pos_embed
        seqlen, batch_size = y.size(1), y.size(0)
        x = q_pos_embed

        # encoder

        for block in self.blocks_2:
            x = block(mask=0, query=x, key=x, values=y,
                      apply_pos=True)  # True: +FFN+残差+laynorm 非第一层与0~t-1的的q的attention, 对应图中Knowledge Retriever
            # mask=0，不能看到当前的response, 在Knowledge Retrever的value全为0，因此，实现了第一题只有question信息，无qa信息的目的
            # print(x[0,0,:])
        return x


class TransformerLayer(nn.Module):
    def __init__(self, d_model, d_feature,
                 d_ff, n_heads, dropout, kq_same, device):
        super().__init__()
        """
            This is a Basic Block of Transformer paper. It containts one Multi-head attention object. Followed by layer norm and postion wise feedforward net and dropout layer.
        """
        kq_same = kq_same == 1
        self.device = device
        # Multi-Head Attention Block
        self.masked_attn_head = MultiHeadAttention(
            d_model, d_feature, n_heads, dropout, kq_same=kq_same, device=device)

        # Two layer norm layer and two droput layer
        self.layer_norm1 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)

        self.linear1 = nn.Linear(d_model, d_ff)
        self.activation = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(d_ff, d_model)

        self.layer_norm2 = nn.LayerNorm(d_model)
        self.dropout2 = nn.Dropout(dropout)

    def forward(self, mask, query, key, values, apply_pos=True):
        """
        Input:
            block : object of type BasicBlock(nn.Module). It contains masked_attn_head objects which is of type MultiHeadAttention(nn.Module).
            mask : 0 means, it can peek only past values. 1 means, block can peek only current and pas values
            query : Query. In transformer paper it is the input for both encoder and decoder
            key : Keys. In transformer paper it is the input for both encoder and decoder
            Values. In transformer paper it is the input for encoder and  encoded output for decoder (in masked attention part)

        Output:
            query: Input gets changed over the layer and returned.

        """

        seqlen, batch_size = query.size(1), query.size(0)
        nopeek_mask = np.triu(
            np.ones((1, 1, seqlen, seqlen)), k=mask).astype('uint8')
        src_mask = (torch.from_numpy(nopeek_mask) == 0).to(self.device)
        if mask == 0:  # If 0, zero-padding is needed.
            # Calls block.masked_attn_head.forward() method
            query2 = self.masked_attn_head(
                query, key, values, mask=src_mask,
                zero_pad=True)  # 只能看到之前的信息，当前的信息也看不到，此时会把第一行score全置0，表示第一道题看不到历史的interaction信息，第一题attn之后，对应value全0
        else:
            # Calls block.masked_attn_head.forward() method
            query2 = self.masked_attn_head(
                query, key, values, mask=src_mask, zero_pad=False)

        query = query + self.dropout1((query2))  # 残差1
        query = self.layer_norm1(query)  # layer norm
        if apply_pos:
            query2 = self.linear2(self.dropout(  # FFN
                self.activation(self.linear1(query))))
            query = query + self.dropout2((query2))  # 残差
            query = self.layer_norm2(query)  # lay norm
        return query


class MultiHeadAttention(nn.Module):
    def __init__(self, d_model, d_feature, n_heads, dropout, kq_same, device, bias=True):
        super().__init__()
        """
        It has projection layer for getting keys, queries and values. Followed by attention and a connected layer.
        """
        self.d_model = d_model
        self.d_k = d_feature
        self.h = n_heads
        self.kq_same = kq_same
        self.device = device

        self.v_linear = nn.Linear(d_model, d_model, bias=bias)
        self.k_linear = nn.Linear(d_model, d_model, bias=bias)
        if kq_same is False:
            self.q_linear = nn.Linear(d_model, d_model, bias=bias)
        self.dropout = nn.Dropout(dropout)
        self.proj_bias = bias
        self.out_proj = nn.Linear(d_model, d_model, bias=bias)

        self._reset_parameters()

    def _reset_parameters(self):
        xavier_uniform_(self.k_linear.weight)
        xavier_uniform_(self.v_linear.weight)
        if self.kq_same is False:
            xavier_uniform_(self.q_linear.weight)

        if self.proj_bias:
            constant_(self.k_linear.bias, 0.)
            constant_(self.v_linear.bias, 0.)
            if self.kq_same is False:
                constant_(self.q_linear.bias, 0.)
            constant_(self.out_proj.bias, 0.)

    def forward(self, q, k, v, mask, zero_pad):

        bs = q.size(0)

        # perform linear operation and split into h heads

        k = self.k_linear(k).view(bs, -1, self.h, self.d_k)
        if self.kq_same is False:
            q = self.q_linear(q).view(bs, -1, self.h, self.d_k)
        else:
            q = self.k_linear(q).view(bs, -1, self.h, self.d_k)
        v = self.v_linear(v).view(bs, -1, self.h, self.d_k)

        # transpose to get dimensions bs * h * sl * d_model

        k = k.transpose(1, 2)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)
        # calculate attention using function we will define next
        scores = attention(q, k, v, self.d_k,
                           mask, self.dropout, zero_pad, self.device)

        # concatenate heads and put through final linear layer
        concat = scores.transpose(1, 2).contiguous() \
            .view(bs, -1, self.d_model)

        output = self.out_proj(concat)

        return output


def attention(q, k, v, d_k, mask, dropout, zero_pad, device):
    """
    This is called by Multi-head atention object to find the values.
    """
    # d_k: 每一个头的dim
    scores = torch.matmul(q, k.transpose(-2, -1)) / \
             math.sqrt(d_k)  # BS, 8, seqlen, seqlen
    bs, head, seqlen = scores.size(0), scores.size(1), scores.size(2)

    scores.masked_fill_(mask == 0, -1e32).to(device)
    scores = F.softmax(scores, dim=-1)  # BS,8,seqlen,seqlen
    # print(f"before zero pad scores: {scores.shape}")
    # print(zero_pad)
    if zero_pad:
        pad_zero = torch.zeros(bs, head, 1, seqlen).to(device)
        scores = torch.cat([pad_zero, scores[:, :, 1:, :]], dim=2)  # 第一行score置0
    # print(f"after zero pad scores: {scores}")
    scores = dropout(scores)
    output = torch.matmul(scores, v)
    # import sys
    # sys.exit()
    return output


class LearnablePositionalEmbedding(nn.Module):
    def __init__(self, d_model, max_len=512):
        super().__init__()
        # Compute the positional encodings once in log space.
        pe = 0.1 * torch.randn(max_len, d_model)
        pe = pe.unsqueeze(0)
        self.weight = nn.Parameter(pe, requires_grad=True)

    def forward(self, x):
        return self.weight[:, :x.size(1), :]  # ( 1,seq,  Feature)


class CosinePositionalEmbedding(nn.Module):
    def __init__(self, d_model, max_len=512):
        super().__init__()
        # Compute the positional encodings once in log space.
        pe = 0.1 * torch.randn(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1).float()
        div_term = torch.exp(torch.arange(0, d_model, 2).float() *
                             -(math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.weight = nn.Parameter(pe, requires_grad=False)

    def forward(self, x):
        return self.weight[:, :x.size(1), :]  # ( 1,seq,  Feature)



import numpy as np
from ..common import EduDataTPL


class EERNNDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ['M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats','M2C_RandomDataSplit4KT', 'M2C_EERNN_OP'],
    }
    
    def get_extra_data(self, **kwargs):
        super_dic = super().get_extra_data(**kwargs)
        super_dic['w2v_word_emb'] = self.w2v_word_emb
        super_dic['exer_content'] = self.content_mat
        return super_dic
    
    def set_info_for_fold(self, fold_id):
        dt_info = self.datatpl_cfg['dt_info']
        dt_info['word_count'] = len(self.word_emb_dict_list[fold_id])

        self.w2v_word_emb = np.vstack(
            [self.word_emb_dict_list[fold_id][k] for k in range(self.datatpl_cfg['dt_info']['word_count'])]
        )

        self.content_mat = self.content_mat_list[fold_id]

    def save_cache(self):
        super().save_cache()
        fph1 = f"{self.cache_folder_path}/word_emb_dict_list.pkl"
        fph2 = f"{self.cache_folder_path}/content_mat_list.pkl"
        self.save_pickle(fph1, self.word_emb_dict_list)
        self.save_pickle(fph2, self.content_mat_list)

    def load_cache(self):
        super().load_cache()
        fph1 = f"{self.cache_folder_path}/word_emb_dict_list.pkl"
        fph2 = f"{self.cache_folder_path}/content_mat_list.pkl"
        self.word_emb_dict_list = self.load_pickle(fph1)
        self.content_mat_list = self.load_pickle(fph2)



from ..common import EduDataTPL
import json
import numpy as np


class RCDDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': [
            'M2C_Label2Int', 'M2C_FilterRecords4CD', 'M2C_ReMapId', 
            'M2C_RandomDataSplit4CD', 'M2C_BuildKCRelation', 
            'M2C_GenQMat', 'M2C_RCD_OP'
        ],
        'M2C_BuildKCRelation': {
            'relation_type': 'rcd_transition',
            'threshold': None
        }
    }

    def get_extra_data(self):
        extra_dict = super().get_extra_data()
        extra_dict['local_map'] = self.local_map
        return extra_dict

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        self.local_map = self.final_kwargs['local_map']
        self.local_map['u_from_e'] = self.local_map['u_from_e_list'][fold_id]
        self.local_map['e_from_u'] = self.local_map['e_from_u_list'][fold_id]



r"""
CDMFKC
##################################
Reference:
    Li et al. "Cognitive Diagnosis Focusing on Knowledge Concepts." in CIKM 2022.
"""
import torch
import torch.nn as nn
from ..utils.components import PosMLP
import torch.nn.functional as F
from ..gd_basemodel import GDBaseModel


class CDMFKC(GDBaseModel):
    """
    dnn_units: dimensions of the middle layers of a multilayer perceptron
    dropout_rate: dropout rate of a multilayer perceptron
    activation: activation function of a multilayer perceptron
    g_impact_a: hyperparameters of the original formula 5
    g_impact_b: hyperparameters of the original formula 5
    """
    default_cfg = {
        'dnn_units': [512, 256],
        'dropout_rate': 0.5,
        'activation': 'sigmoid',
        'g_impact_a': 0.5,
        'g_impact_b': 0.5
    }
    def __init__(self, cfg):
        """Pass parameters from other templates into the model

        Args:
            cfg (UnifyConfig): parameters from other templates
        """
        super().__init__(cfg)

    def build_cfg(self):
        """Initialize the parameters of the model"""
        self.n_user = self.datatpl_cfg['dt_info']['stu_count']
        self.n_item = self.datatpl_cfg['dt_info']['exer_count']
        self.n_cpt = self.datatpl_cfg['dt_info']['cpt_count']

    def add_extra_data(self, **kwargs):
        """Add the data required by the model to the self object from the data template"""
        self.Q_mat = kwargs['Q_mat'].to(self.device)

    def build_model(self):
        """Initialize the various components of the model"""
        self.student_emb = nn.Embedding(self.n_user, self.n_cpt)
        self.k_difficulty = nn.Embedding(self.n_item, self.n_cpt)
        self.e_difficulty = nn.Embedding(self.n_item, 1)
        self.k_impact = nn.Embedding(self.n_item, self.n_cpt)
        self.pd_net = PosMLP(
            input_dim=self.n_cpt, output_dim=1, activation=self.modeltpl_cfg['activation'],
            dnn_units=self.modeltpl_cfg['dnn_units'], dropout_rate=self.modeltpl_cfg['dropout_rate']
        )

    def forward(self, stu_id, exer_id, **kwargs):
        """Get the probability that the students will answer the exercise correctly

        Args:
            stu_id (torch.Tensor): Id of students. Shape of [inter_len]
            exer_id (torch.Tensor): Sequence of exercise id. Shape of [inter_len]

        Returns:
            torch.Tensor: prediction_tensor
        """
        # before prednet
        items_Q_mat = self.Q_mat[exer_id]  # Q_mat: exer_num * n_cpt; items_Q_mat: batch_exer_num * n_cpt
        stu_emb = self.student_emb(stu_id)
        stat_emb = torch.sigmoid(stu_emb)

        k_difficulty = torch.sigmoid(self.k_difficulty(exer_id))
        e_difficulty = torch.sigmoid(self.e_difficulty(exer_id))
        h_impact = torch.sigmoid(self.k_impact(exer_id))
        g_impact = torch.sigmoid(self.modeltpl_cfg['g_impact_a'] * h_impact + 
                                 self.modeltpl_cfg['g_impact_b'] * k_difficulty * e_difficulty)

        input_knowledge_point = items_Q_mat
        input_x = e_difficulty * (stat_emb + g_impact - k_difficulty) * input_knowledge_point

        pd = self.pd_net(input_x).sigmoid()

        return pd


    def get_main_loss(self, **kwargs):
        """Get the loss in the paper

        Returns:
            dict: {'loss_main': loss_value}
        """
        stu_id = kwargs['stu_id']
        exer_id = kwargs['exer_id']
        label = kwargs['label']
        pd = self(stu_id, exer_id).flatten()
        loss = F.binary_cross_entropy(input=pd, target=label)
        return {
            'loss_main': loss
        }

    def get_loss_dict(self, **kwargs):
        """

        Returns:
            dict: loss dict{'loss_main': loss_value}
        """
        return self.get_main_loss(**kwargs)
    
    @torch.no_grad()
    def predict(self, stu_id, exer_id,  **kwargs):
        """A function of get how well the model predicts students' responses to exercise questions

        Args:
            stu_id (torch.Tensor): Id of students.
            exer_id (torch.Tensor): Id of exercise.

        Returns:
            dict: {'y_pd':prediction_tensor}
        """
        return {
            'y_pd': self(stu_id, exer_id).flatten(),
        }
    
    def get_stu_status(self, stu_id=None):
        """Get the Cognitive State of the Student

        Args:
            stu_id (torch.Tensor, optional): Id of students. Defaults to None.

        Returns:
            nn.Embedding: Cognitive State of the Student
        """
        if stu_id is not None:
            return self.student_emb(stu_id)
        else:
            return self.student_emb.weight



from ..common import EduDataTPL
import numpy as np
import torch


class GKTDataTPL(EduDataTPL):
    default_cfg = {
        'mid2cache_op_seq': ["M2C_KCAsExer", 'M2C_Label2Int', 'M2C_ReMapId', 'M2C_BuildSeqInterFeats', 'M2C_RandomDataSplit4KT'],
    }

    def process_load_data_from_middata(self):
        super().process_load_data_from_middata()
        concept_num = self.final_kwargs['dt_info']['exer_count']
        if self.modeltpl_cfg['graph_type'] == 'Dense':
            self.final_kwargs['gkt_graph'] = self.build_dense_graph(node_num=concept_num)
        elif self.modeltpl_cfg['graph_type'] == 'Transition':
            self.final_kwargs['gkt_graph_list'] = []
            for train_dict in self.final_kwargs['df_train_folds']:
                self.final_kwargs['gkt_graph_list'].append(
                    self.build_transition_graph(
                        exer_seq=train_dict['exer_seq:token_seq'],
                        mask_seq=train_dict['mask_seq:token_seq'],
                        concept_num=concept_num
                    )
                )   
        elif self.modeltpl_cfg['graph_type'] == 'DKT':
            self.final_kwargs['gkt_graph'] = self.build_dkt_graph(
                f"{self.frame_cfg.data_folder_path}/dkt_graph.txt", 
                concept_num=concept_num
            )
        elif self.modeltpl_cfg['graph_type'] == 'PAM':
            self.final_kwargs['gkt_graph'] = None
        elif self.modeltpl_cfg['graph_type'] ==  'MHA':
            self.final_kwargs['gkt_graph'] = None
        elif self.modeltpl_cfg['graph_type'] ==  'VAE':
            self.final_kwargs['gkt_graph'] = None
        else:
            raise ValueError(f"unknown graph_type: {self.modeltpl_cfg['graph_type']}")

    def get_extra_data(self, **kwargs):
        dic = super().get_extra_data(**kwargs)
        dic['graph'] = self.final_kwargs['gkt_graph']
        return dic
    
    @staticmethod
    def build_transition_graph(exer_seq, mask_seq, concept_num):
        graph = np.zeros((concept_num, concept_num))
        for exers, masks in zip(exer_seq, mask_seq):
            for idx in range(len(exers) - 1):
                if masks[idx+1] == 0: break
                graph[exers[idx], exers[idx+1]] += 1
        np.fill_diagonal(graph, 0)
        # row normalization
        rowsum = np.array(graph.sum(1))
        def inv(x):
            if x == 0:
                return x
            return 1. / x
        inv_func = np.vectorize(inv)
        r_inv = inv_func(rowsum).flatten()
        r_mat_inv = np.diag(r_inv)
        graph = r_mat_inv.dot(graph)
        # covert to tensor
        graph = torch.from_numpy(graph).float()
        return graph
    
    @staticmethod
    def build_dense_graph(node_num):
        graph = 1. / (node_num - 1) * np.ones((node_num, node_num))
        np.fill_diagonal(graph, 0)
        graph = torch.from_numpy(graph).float()
        return graph

    @staticmethod
    def build_dkt_graph(file_path, concept_num):
        graph = np.loadtxt(file_path)
        assert graph.shape[0] == concept_num and graph.shape[1] == concept_num
        graph = torch.from_numpy(graph).float()
        return graph

    def set_info_for_fold(self, fold_id):
        super().set_info_for_fold(fold_id)
        if self.modeltpl_cfg['graph_type'] == 'Transition':
            self.final_kwargs['gkt_graph'] = self.final_kwargs['gkt_graph_list'][fold_id]



from .base_mid2cache import BaseMid2Cache
import pandas as pd
import numpy as np
from itertools import chain


class M2C_BuildKCRelation(BaseMid2Cache):
    default_cfg = {
        'relation_type': 'rcd_transition',
        'threshold': None
    }

    def process(self, **kwargs):
        df = kwargs['df']
        if df is None:
            df = pd.concat(
                [kwargs['df_train'], kwargs['df_valid'], kwargs['df_test']],
                axis=0, ignore_index=True
            )

        if 'order_id:float' in df:
            df.sort_values(by=['order_id:float'], axis=0, ignore_index=True, inplace=True)
        elif 'order_id:token' in df:
            df.sort_values(by=['order_id:token'], axis=0, ignore_index=True, inplace=True)
        
        kwargs['df'] = df

        if self.m2c_cfg['relation_type'] == 'rcd_transition':
            return self.gen_rcd_transition_relation(kwargs)
        else:
            raise NotImplementedError
        
    def gen_rcd_transition_relation(self, kwargs):
        df:pd.DataFrame = kwargs['df']
        df_exer = kwargs['df_exer'][['exer_id:token', 'cpt_seq:token_seq']]
        df = df[['stu_id:token', 'exer_id:token', 'label:float']].merge(df_exer[['exer_id:token', 'cpt_seq:token_seq']], how='left', on='exer_id:token')
        cpt_count = np.max(list(chain(*df_exer['cpt_seq:token_seq'].to_list()))) + 1
        
        n_mat = np.zeros((cpt_count, cpt_count), dtype=np.float32) # n_{i,j}
        for _, df_one_stu in df.groupby('stu_id:token'):
            for idx in range(df_one_stu.shape[0]):
                if idx == df_one_stu.shape[0] - 2:
                    break
                curr_record = df_one_stu.iloc[idx]
                next_record = df_one_stu.iloc[idx + 1]
                if curr_record['label:float'] * next_record['label:float'] == 1:
                    for cpt_pre in curr_record['cpt_seq:token_seq']:
                        for cpt_next in next_record['cpt_seq:token_seq']:
                            if cpt_pre != cpt_next:
                                n_mat[cpt_pre, cpt_next] += 1

        a = np.sum(n_mat, axis=1)[:,None]
        nonzero_mask = (a != 0)
        np.seterr(divide='ignore', invalid='ignore')
        C_mat = np.where(nonzero_mask, n_mat / a, n_mat)

        max_val = C_mat.max()
        np.fill_diagonal(C_mat, max_val)
        min_val = C_mat.min()
        np.fill_diagonal(C_mat, 0)
        T_mat = (C_mat- min_val) / (max_val - min_val)

        threshold = self.m2c_cfg['threshold']
        if threshold is None:
            threshold = C_mat.sum() / (C_mat != 0).sum()
            threshold *= threshold
            threshold *= threshold
        
        cpt_dep_mat = (T_mat > threshold).astype(np.int64)

        kwargs['cpt_dep_mat'] = cpt_dep_mat

        return kwargs



# -*- coding: utf-8 -*-
# @Author : Xiangzhi Chen
# @Github : kervias

import os
import sys
import time
from functools import wraps
import pytz
import datetime
import json
import random
import torch
import numpy as np
import logging


tensor2npy = lambda x: x.cpu().detach().numpy() if x.is_cuda else x.detach().numpy()
tensor2cpu = lambda x: x.cpu() if x.is_cuda else x


def set_same_seeds(seed):
    # Python built-in random module
    random.seed(seed)
    # Numpy
    np.random.seed(seed)
    # Torch
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


class PathUtil(object):
    @staticmethod
    def auto_create_folder_path(*args):
        for path in args:
            if not os.path.exists(path):
                os.makedirs(path)

    @staticmethod
    def get_main_folder_path():
        try:
            __IPYTHON__
        except NameError:
            return os.path.realpath(os.path.dirname(os.path.abspath(sys.argv[0])))
        return os.getcwd()

    @staticmethod
    def check_path_exist(*args):
        for path in args:
            if not os.path.exists(path):
                raise FileNotFoundError(os.path.realpath(path) + " not exists")


class IOUtil(object):
    @staticmethod
    def read_json_file(filepath, encoding='utf-8'):
        with open(filepath, 'r', encoding=encoding) as f:
            return json.load(f)

    @staticmethod
    def write_json_file(filepath, data, encoding='utf-8', ensure_ascii=False, indent=2):
        with open(filepath, 'w', encoding=encoding) as f:
            json.dump(data, f, ensure_ascii=ensure_ascii, indent=indent)


class IDUtil(object):
    @staticmethod
    def get_random_id_bytime():
        tz = pytz.timezone('Asia/Shanghai')
        return datetime.datetime.now(tz).strftime("%Y-%m-%d-%H%M%S")


def get_gpu_usage(device=None):
    r""" Return the reserved memory and total memory of given device in a string.
    Args:
        device: cuda.device. It is the device that the model run on.
    Returns:
        str: it contains the info about reserved memory and total memory of given device.
    """

    reserved = torch.cuda.max_memory_reserved(device) / 1024 ** 3
    total = torch.cuda.get_device_properties(device).total_memory / 1024 ** 3

    return '{:.2f} G/{:.2f} G'.format(reserved, total)


class DecoratorTimer:
    def __init__(self):
        self.logger = logging.getLogger("edustudio")

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.logger.info(f'Function:{func.__name__} start running...')
            start_time = time.time()
            temp = func(*args, **kwargs)
            end_time = time.time()
            self.logger.info(f'Function:{func.__name__} running time: {end_time - start_time:.4f} sec')
            return temp

        return wrapper